import numpy as np
from tqdm import trange

import modules.scripts as scripts
import gradio as gr

#from PIL import ImageFilter
from PIL import Image, ImageDraw, ImageOps, ImageFilter

from modules import processing, shared, sd_samplers, images
from modules.processing import Processed
from modules.sd_samplers import samplers
from modules.shared import opts, cmd_opts, state

import random
from glob import glob
import copy

import cv2

class Script(scripts.Script):
	def title(self):
		return "Double Diffusion Ver .04"

	def show(self, is_img2img):
		return is_img2img

	def ui(self, is_img2img):
		numcircles = gr.Slider(minimum=0, maximum=1000, step=1, label='Number of Alpha Circles', value=10)
		minradw = gr.Slider(minimum=1, maximum=500, step=1, label='Min radius width', value=60)
		maxradw = gr.Slider(minimum=1, maximum=500, step=1, label='Max radius width', value=120)
		minradh = gr.Slider(minimum=1, maximum=500, step=1, label='Min radius height', value=8)
		maxradh = gr.Slider(minimum=1, maximum=500, step=1, label='Max radius height', value=30)
		minalpha = gr.Slider(minimum=0, maximum=1, step=.01, label='Min Alpha', value=.8)
		maxalpha = gr.Slider(minimum=0, maximum=1, step=.01, label='Max Alpha', value=1.)
		blurrad = gr.Slider(minimum=0.0, maximum=100.0, step=0.1, label='Blur Radius', value=17.0)
		eyeprob = gr.Slider(minimum=0.0, maximum=1.0, step=0.1, label='Eye preservation probability', value=.8)
		mouthprob = gr.Slider(minimum=0.0, maximum=1.0, step=0.1, label='Mouth preservation probability', value=.6)
		second_denoise_strength = gr.Slider(minimum=0.0, maximum=1.0, step=0.01, label='Second Pass Denoising Strength', value=.4)
		secondsteps = gr.Slider(minimum=1, maximum=150, step=1, label='Second Pass Steps', value=100)

		return [numcircles, minradw, maxradw, minradh, maxradh, minalpha, maxalpha, blurrad, eyeprob, mouthprob, second_denoise_strength, secondsteps]

	def run(self, p, numcircles, minradw, maxradw, minradh, maxradh, minalpha, maxalpha, blurrad, eyeprob, mouthprob, second_denoise_strength, secondsteps):
		face_cascade = cv2.CascadeClassifier(cv2.data.haarcascades + 'haarcascade_frontalface_default.xml')
		eye_cascade = cv2.CascadeClassifier(cv2.data.haarcascades + 'haarcascade_eye.xml')
		mouth_cascade = cv2.CascadeClassifier("C:/ai/haarcascade_mcs_mouth.xml")

		def blend_two(im, secondimg):
			# Load the image and apply the Canny edge detection algorithm
			orig_im = im
			blend_image = secondimg
			blend_image = blend_image.resize(im.size, Image.ANTIALIAS)

			#---fun stuff starts here---
			pil_image_array = np.array(im)
			img = cv2.cvtColor(pil_image_array, cv2.COLOR_RGB2BGR)
			gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
			#set image to all white
			img[:] = 255
			gray_face = gray

			image = np.array(blend_image)
			gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
			edges = cv2.Canny(gray_face, 10, 75, 7, L2gradient = True)
			kwidth, kheight = edges.shape
			kwidth *= .1
			ksize = (int(kwidth), int(kwidth))
			blursize = (int(blurrad), int(blurrad))
			#cv2.imshow("edges", edges)

			# Calculate the gradient of the edges using the Sobel operator
			gradient_x = cv2.Sobel(edges, cv2.CV_64F, 1, 0, ksize=3)
			gradient_y = cv2.Sobel(edges, cv2.CV_64F, 0, 1, ksize=3)

			gradient_x = cv2.blur(gradient_x, ksize)
			gradient_y = cv2.blur(gradient_y, ksize)

			# Calculate the magnitude and angle of the gradient
			magnitude, angle = cv2.cartToPolar(gradient_x, gradient_y, angleInDegrees=True)
			#create correction image containing solid 180 for modulo hack
			correction_img = np.zeros(gray.shape, np.uint8)
			correction_img.fill(180.0)
			modangle = np.mod(angle, correction_img)

			#draw ellipses
			ellis = np.zeros(modangle.shape)
			height, width = modangle.shape

			#Draw n randomly placed and sized ellipses
			for i in range(int(numcircles)):
				# Generate random center point, axes lengths, and angle of rotation for the ellipse
				center = (np.random.randint(0, width), np.random.randint(0, height))
				axes = (np.random.randint(int(minradw), int(maxradw)), np.random.randint(int(minradh), int(maxradh)))
				cx = center[0]
				cy = center[1]
				angle = modangle[cy, cx]
				colval = random.uniform(minalpha, maxalpha)

				# Draw the ellipse onto the image
				cv2.ellipse(ellis, center, axes, angle + 90., 0, 360, colval, -1)

			for i in range(int(numcircles * .12)):
				# Generate random center point, axes lengths, and angle of rotation for the ellipse
				center = (np.random.randint(0, width), np.random.randint(0, height))
				axes = (np.random.randint(int(minradw*3), int(maxradw*3)), np.random.randint(int(minradh*3), int(maxradh*3)))
				cx = center[0]
				cy = center[1]
				angle = modangle[cy, cx]
				colval = random.uniform(minalpha, maxalpha)

				# Draw the ellipse onto the image
				cv2.ellipse(ellis, center, axes, angle + 90., 0, 360, colval, -1)

			# Draw ellipses around eyes
			eyes = eye_cascade.detectMultiScale(gray_face, 1.3, 5)
			for (ex, ey, ew, eh) in eyes:
				if random.random() < eyeprob:
					eyectr = (int(ex + ew * .5), int(ey + eh * .5))
					angle = modangle[eyectr]
					cv2.ellipse(ellis, (ex + ew // 2, ey + eh // 2), (ew // 2, eh // 3), angle, 0, 360, 1., -1)

			# Draw ellipses around mouth(s)
			mouths = mouth_cascade.detectMultiScale(gray_face, 1.3, 5)
			for (ex, ey, ew, eh) in mouths:
				if random.random() < mouthprob:
					mouthctr = (int(ex + ew * .5), int(ey + eh * .5))
					cv2.ellipse(ellis, (ex + ew // 2, ey + eh // 2), (ew // 2, eh // 3), 0, 0, 360, 1., -1)

			# Display the image
			ellis = cv2.blur(ellis, blursize)
			whiteimg = np.ones(ellis.shape)
			ellis = np.subtract(whiteimg, ellis)
			ellis = np.multiply(ellis, 255.0)

			tempimg = Image.fromarray(ellis)
			mask = tempimg.convert("L")
			alpha = mask.convert("RGBA")

			im.format = blend_image.format
			blend_image.putalpha(mask)
			im.putalpha(255)

			outimg = Image.alpha_composite(im, blend_image)
			outimg.format = orig_im.format
			
			return outimg


		processing.fix_seed(p)
		batch_count = p.n_iter
		p.extra_generation_params = {
			"Number of Circles": numcircles,
			"Min Radius Width": minradw,
			"Max Radius Width": maxradw,
			"Min Radius Height": minradh,
			"Max Radius Height": maxradh,
			"Min Alpha": minalpha,
			"Max Alpha": maxalpha,
			"Blur Radius": blurrad,
			"Eye Probability": eyeprob,
			"Mouth Probability": mouthprob,
			"Second Pass Denoise Strength": second_denoise_strength,
			"Second Pass Steps": secondsteps
		}

		p.batch_size = 1
		p.n_iter = 1

		output_images, info = None, None
		initial_seed = None
		initial_info = None

		grids = []
		all_images = []
		
		state.job_count = batch_count * 2
		initial_steps = p.steps
		initial_denoising_strength = p.denoising_strength
		initial_color_corrections = [processing.setup_color_correction(p.init_images[0])]
		original_init_image = copy.deepcopy(p.init_images)
		initial_imgs = []
		random.seed(p.seed)

		#--------
		for n in range(batch_count):
			history = []
			p.steps = initial_steps
			# Reset to original init image at the start of each batch
			p.init_images = original_init_image
			p.denoising_strength = initial_denoising_strength
			initial_imgs.append(original_init_image[0])

			p.n_iter = 1
			p.batch_size = 1
			p.do_not_save_grid = True

			print("Initial denoising strength:")
			print(p.denoising_strength)

			if opts.img2img_color_correction:
				p.color_corrections = initial_color_corrections

			processed = processing.process_images(p)

			if initial_seed is None:
				initial_seed = processed.seed
				initial_info = processed.info

			p.seed = processed.seed + 1
			processed.images = processed.images[0]

			history.append(processed.images)

			initme = []
			initimg = blend_two(initial_imgs[0], processed.images)
			initme.append(initimg)
			p.init_images = initme
			p.denoising_strength = second_denoise_strength

			p.steps = secondsteps
			processed = processing.process_images(p)
			if initial_seed is None:
				initial_seed = processed.seed
				initial_info = processed.info

			p.seed = processed.seed + 1
			
			processed.images = processed.images[0]
			history.append(processed.images)

			grid = images.image_grid(history, rows=3)
			if opts.grid_save:
				images.save_image(grid, p.outpath_grids, "grid", initial_seed, p.prompt, opts.grid_format, info=info, short_filename=not opts.grid_extended_filename, grid=True, p=p)

			grids.append(grid)
			all_images += history
			p.init_images = original_init_image

		if opts.return_grid:
			all_images = grids + all_images


		processed = Processed(p, all_images, initial_seed, initial_info)

		return processed